@@ -1,15 +1,17 @@
 # -----------------------------------------------------------------------------
 # IMPORTS
 # -----------------------------------------------------------------------------
-import sys
-sys.path.append('../')
+from inspect import signature
 
-from abc import ABC, abstractproperty
+import pandas as pd
+
+from abc import ABC, abstractproperty, abstractclassmethod
 import datetime
 from re import compile
 from collections import OrderedDict
 from os import popen
 import os
+import threading
 
 intBeginWith = compile(r'^\D+')
 
@@ -18,13 +20,18 @@ intBeginWith = compile(r'^\D+')
 # -----------------------------------------------------------------------------
 
 class Default():
-	"""Default docString"""
+	"""Default class representing class docString template"""
 	def __str__(self): return self.__doc__
+	def _repr(self):
+		fields = signature(self.__init__).parameters
+		values = ", ".join(repr(getattr(self, f)) for f in fields)
+		return f'{type(self).__name__}({values})'
+
 
 
 class Container(ABC):
-	"""Support Containers callable objects
-	object should contain objVar property
+	"""Abstract base class providing template for standard dunder 
+	methods template.  Object should contain objVar property
 	"""
 	@abstractproperty
 	@property
@@ -36,7 +43,13 @@ class Container(ABC):
 	def __setitem__(self, i, v): self.objVar[i] = v
 	def __delitem__(self, i): del(self.objVar[i])
 	def __contains__(self, i): return i in self.objVar
-	def __reversed__(self): return reversed(self.objVar)
+	def __reversed__(self): 
+		if isinstance(self.objVar, dict):
+			for k in reversed(tuple(self.objVar.keys())):
+				yield (k,self[k])
+		else:
+			return reversed(self.objVar)
+		
 	def __missing__(self, i): raise Exception(f'key {i} unavailable') # only for dict subclass
 	def __iter__(self):
 		if isinstance(self.objVar, (list, tuple, set, str)):
@@ -83,6 +96,9 @@ class Numeric():
 #                           STRING OPERATIONS                                 #
 # -----------------------------------------------------------------------------
 class STR(Container):
+	"""Collection of static methods for string objects.
+	see more...	
+	"""
 
 	@staticmethod
 	def foundPos(s, sub, pos=0):
@@ -378,7 +394,9 @@ class STR(Container):
 
 	@staticmethod
 	def hostname(net_connect):
-		'''Hostname from connection'''
+		'''input paramiko netconnection, returns hostname from device.
+		-->str (Hostname from connection)
+		'''
 		try:
 			hns = net_connect.find_prompt()[:-1]
 			atPos = STR.foundPos(hns, "@")
@@ -389,7 +407,9 @@ class STR(Container):
 
 	@staticmethod
 	def hostname_from_cli(line, command):
-		'''hostname from command line'''
+		'''input standard text input line, for which command was entered.
+		-->str (hostname from command line)
+		'''
 		if not STR.found(line, command): return None
 		cmdPos = STR.foundPos(line, command)
 		hn = line[:cmdPos].strip()[:-1]
@@ -397,24 +417,41 @@ class STR(Container):
 
 	@staticmethod
 	def shrink_if(intName, length=2):
-		'''Interface Name shortening for standard 2 Characters '''
+		'''Interface Name shortening, input length will decide number of 
+		charactes to be included in shortened output
+		-->str (short name of interface)
+		'''
 		iBW = intBeginWith.match(intName)
 		return iBW.group()[:length]+intName[iBW.span()[1]:]
 
 	@staticmethod
 	def if_prefix(intName):
-		'''Interface beginning Name'''
+		'''Interface beginning Name
+		-->str (interface prefix)
+		'''
 		iBW = intBeginWith.match(intName)
 		return intName[iBW.start(): iBW.end()]
 
 	@staticmethod
 	def update_str(s, searchItem='', replaceItem=''):
-		'''Updates line for search item with replace item'''
+		'''Updates line for search item with replace item
+		(Find/Repalace)
+		-->str
+		'''
 		return s.replace(searchItem, replaceItem)
 
 	@staticmethod
-	def get_logfile_name(folder, hn, cmd='', ts='', separator="_@_"):
-		'''return log file name for the command on device with/wo provided time_stamp'''
+	def get_logfile_name(folder, hn, cmd='', ts='', separator="_@_", extn='.log'):
+		'''return log file name for the command on device with/wo provided 
+		time_stamp.
+		folder = path to where file should be saved
+		hn = file name starting with provided host-name
+		cmd = file name containing additional commands string
+		ts = file name containing additional time stamp
+		separator = hn-cmd-ts separator 
+		extn = extension of filename
+		--> str (filename along with full path)
+		'''
 		if ts: ts = separator + ts
 		if cmd:
 			cmd += ts
@@ -422,11 +459,12 @@ class STR(Container):
 			for x in replaceCandidates:
 				cmd = STR.update_str(cmd, x, "_")
 			cmd = separator + cmd
-		return folder+hn+cmd+'.log'
+		return folder+hn+cmd+extn
 
 	@staticmethod
 	def string_concate(s, s1, conj=''):
 		'''Concatenate strings s and s1 with conjuctor conj
+		--> str
 
 		:param s: string
 		:type s: str
@@ -437,11 +475,7 @@ class STR(Container):
 		:param conj: conjuctor
 		:type conj: string
 		'''
-		if s == '': 
-			s = s + s1
-		else:
-			s = s + conj + s1
-		return s
+		return s + s1 if s == '' else s + conj + s1
 
 	@staticmethod
 	def right(strg, n):
@@ -494,16 +528,37 @@ class STR(Container):
 
 	@staticmethod
 	def to_list(s):
-		'''Returns list for the provided string - s, split by lines '''
+		'''Returns list for the provided string - s, 
+		splits string by lines
+		--> list
+		'''
 		s = s.split("\n")
 		for i, x in enumerate(s):
 			s[i] = x + "\n"
 		return s
-		# return s.split("\n")
+
+	@staticmethod
+	def to_set(s):
+		'''Return set of values for the provided string - s.
+		splits string by lines and comma
+		--> set
+		'''
+		if isinstance(s, str):
+			_s = []
+			for _ in s.split('\n'):
+				_s.extend(_.split(','))
+			return set(LST.remove_empty_members((_s)))
+		else:
+			return set(s)
 
 	@staticmethod
 	def header_indexes(line):
-		exceptional_headers = {'Type', }
+		"""input header string line of a text table.
+		returns dictionary with key:value pair where 
+		keys are header string and value are string index (position) of string in line
+		--> OrderedDict
+		"""
+		exceptional_headers = {}#'Type', }
 		headers = OrderedDict()
 		prev_k = None
 		for k in STR.replace_dual_and_split(line.rstrip()):
@@ -519,24 +574,52 @@ class STR(Container):
 
 	@staticmethod
 	def prepend_bgp_as(bgp_as, n):
+		"""'n' number of BGP AS Number prepending string.
+		--> str
+		"""
 		s = ''
 		for x in range(n): s += str(bgp_as) + " "
 		return s[:-1]
 
 	@staticmethod
-	def ending(line, c): return line.strip().endswith(c)
+	def ending(line, c): 
+		"""check if line ends with c or not
+		-->boolean
+		"""
+		return line.strip().endswith(c)
 
 	@staticmethod
-	def starting(line, c): return line.strip().startswith(c)
+	def starting(line, c): 
+		"""check if line starts with c or not
+		-->boolean
+		"""
+		return line.strip().startswith(c)
 
 
 # -----------------------------------------------------------------------------
 #                    FILE OPERATIONS/ CONVERSIONS                             #
 # -----------------------------------------------------------------------------
 class IO():
+	"""Collection of static methods for IO objects.
+	see more...	
+	"""
+
+	@staticmethod
+	def copy_text_file(file):
+		"""copy file.txt to file-copy.txt
+		--> None
+		"""
+		dst = file[:-4] + "-copy.txt"
+		with open(file, 'r') as sf:
+			with open(dst, 'w') as dt:
+				dt.write(sf.read())
 
 	@staticmethod
 	def file_list_for_time_stamp(hn, ts, folder, splitter="_@_" ):
+		"""collection of files from given folder where hostname (hn) 
+		and timestamp (ts) found in the file name.
+		--> set
+		"""
 		files = set()
 		for file in os.listdir(folder):
 			if not splitter in file: continue
@@ -546,6 +629,9 @@ class IO():
 
 	@staticmethod
 	def devices_on_log_files(folder, splitter="_@_"):
+		"""collection of files from given folder where file extensions are .log
+		--> set
+		"""
 		devices = set()
 		for file in os.listdir(folder):
 			if not splitter in file: continue
@@ -556,6 +642,10 @@ class IO():
 
 	@staticmethod
 	def timestamps_for_device(devname, folder, splitter="_@_"):
+		"""collection of time stamps of files from given folder
+		for given hostnames.
+		--> set
+		"""
 		stamps = set()
 		for file in os.listdir(folder):
 			if not splitter in file: continue
@@ -585,6 +675,8 @@ class IO():
 		:param file: text input file name/with path
 		:type file: str
 		'''
+		file = file.strip()
+		if file is None: return None
 		with open(file, 'r') as f:
 			lines = f.readlines()
 		return lines
@@ -633,14 +725,14 @@ class IO():
 		:param cr: carriage return to add at end of each string/line.(default True)
 		:type cr: bool
 		'''
-		if filename != '':
-			if isinstance(matter, str):
-				if cr: matter += "\n" 
-				with open(filename, 'a') as f:
-					f.write(matter)
-			elif isinstance(matter, (list, tuple ,set)):
-				for i in matter:
-					IO.add_to_file(filename, i)
+		if not filename: return None
+		if isinstance(matter, str):
+			if cr and matter and matter[-1] != "\n": matter += "\n"
+			with open(filename, 'a') as f:
+				f.write(matter)
+		elif isinstance(matter, (list, tuple ,set)):
+			for i in matter:
+				IO.add_to_file(filename, i)
 
 	@staticmethod
 	def update(file, find_item, replace_item):
@@ -667,6 +759,10 @@ class IO():
 
 	@staticmethod
 	def jinja_verification(folder):
+		"""check all text files from provided folder for verification of 
+		self jinja strings descrepencies
+		--> str with outcome
+		"""
 		s = ''
 		for file in os.listdir(folder):
 			goahead = {'GOAHEAD FOR': 0, 'GOAHEAD END': 0,}
@@ -688,15 +784,25 @@ class IO():
 # -----------------------------------------------------------------------------
 
 class LST():
+	""" Collection of static methods for list objects.
+	see more...	
+	"""
 
 	@staticmethod
 	def remove_empty_members(lst):
+		"""house keeping of list
+		removes empty members from list
+		-->list
+		"""
 		empty_members = ('', None, 'N/A', 'nil')
 		tmp_lst = [m for m in lst if not m in empty_members]
 		return tmp_lst
 
 	@staticmethod
 	def convert_vlans_list_to_range_of_vlans_list(vlan_list):
+		"""converts list of individual vlans to a list of range of vlans
+		--> list
+		"""
 		vlan_list = sorted(vlan_list)
 		vlan_list.append(None)
 		range_list, previous_vlan = [], 0
@@ -718,10 +824,14 @@ class LST():
 
 	@staticmethod
 	def list_variants(input_list):
+		"""list of vlans in different format
+		list of vlans,
+		space separated string,
+		comma separated string,		
+		--> dict
+		"""
 		str_list = [str(_) 
 			for _ in LST.convert_vlans_list_to_range_of_vlans_list(input_list)]
-		# str_list = [str(_) 
-		# 	for _ in input_list]
 		ssv_list = " ".join(str_list)
 		csv_list = ",".join(str_list)
 		return {
@@ -732,8 +842,12 @@ class LST():
 
 	@staticmethod
 	def list_of_devices(list_of_files):
+		"""get hostnames (first index item) from list of files.
+		--> set
+		"""
 		devices = set()
 		for file in list_of_files:
+			if not file.strip(): continue
 			f = ".".join(os.path.basename(file).split(".")[:-1])
 			hn = f.split("_")[0]
 			if not hn in devices: devices.add(hn)
@@ -741,12 +855,24 @@ class LST():
 
 	@staticmethod
 	def split(lst, n):
+		"""yield provided list with group of n number of items
+		--> generator of list
+		"""
 		s = 0
 		lst = tuple(lst)
 		for _ in range(s, len(lst), n):
 			yield lst[_: s+n]
 			s += n
 
+	@staticmethod
+	def list_to_octet(lst):
+		"""joins and return string with provided list with '.'
+		helpful in created ipv4 string with list of 4 numeric items
+		--> str
+		"""
+		l = ''
+		for x in lst: l = str(x) if l == '' else l +'.'+ str(x)
+		return l
 
 
 # -----------------------------------------------------------------------------
@@ -754,6 +880,9 @@ class LST():
 # -----------------------------------------------------------------------------
 
 class DIC():
+	"""Collection of static methods for dictionary objects.
+	see more...	
+	"""
 
 	# INTERNAL : update dictionary d for provided keyvalue pairs
 	# param: d: dest dictionary
@@ -786,30 +915,162 @@ class DIC():
 
 	@staticmethod
 	def recursive_dic(dic, indention=0):
+		"""convert dictionary (dic) to string. 
+		recursive dictionary increases indention.
+		--> str
+		"""
 		s = ""
 		if isinstance(dic, dict):
 			for k, v in dic.items():
-				s += f"{'  '*indention}{k}\n"
-				if isinstance(v, dict):
-					indention += 1
-					s += DIC.recursive_dic(v, indention)
-					indention -= 1
-				elif isinstance(v, (list,set)):
-					for x in v: 
-						s += x+'\n'
-				else:
-					s += f"  {'  '*indention}{v}\n"
-		elif isinstance(dic, (list,set)):
+				s += f"{' '*indention}{k}\n"
+				indention += 1
+				s += DIC.recursive_dic(v, indention)
+				indention -= 1
+		elif isinstance(dic, (tuple,list,set)):
 			for x in dic: 
-				s += x+'\n'
+				if x: s += str(x)+'\n'
+		elif isinstance(dic, str):
+			if dic: s+= f"  {' '*indention}{dic}\n"
 		return s
 
+# -----------------------------------------------------------------------------
+#                         DICTIONARY DIFFERECES                               #
+# -----------------------------------------------------------------------------
+
+class DifferenceDict(dict):
+	"""Template class to get difference in two dictionary objects.
+	use dunder +/- for adds/removes.	
+	"""
+
+	missing = "- "
+	additive = "+ "
+
+	def __init__(self, d):
+		self.d = d
+
+	def __sub__(self, d): return self.get_change(d, self.missing)
+	def __add__(self, d): return self.get_change(d, self.additive)
+
+	def get_change(self, d, change):
+		"""compare current object/dict with provided new object/dict (ie: d) 
+		and return differences based on change required ("- "/"+ ")
+		"""
+		if isinstance(d, DifferenceDict):
+			return dict_differences(self.d, d.d, change)
+		elif isinstance(d, dict):
+			return dict_differences(self.d, d, change)
+
+# INTERNAL / RECURSIVE
+# returns differences for provided subnet/change
+# input subject can be of string/int/float/set/dictionary
+# input change is change type prefix
+# return value type depends on input subject type.
+def _get_differences(subject, change):
+	if isinstance(subject, (str, int, float)):
+		diff = change + str(subject)
+	elif isinstance(subject, set):
+		diff = set()
+		for item in subject:
+			df = _get_differences(item, change)
+			diff.add(df)
+	elif isinstance(subject, dict):
+		diff = dict()
+		for key, value in subject.items():
+			key = change + str(key)
+			if value:
+				diff[key] = _get_differences(value, change)
+			else: 
+				diff[key] = ''
+	else:
+		raise Exception(f"InvalidSubjectTypeError: {type(subject)}:{subject}")
+	return diff
+
+
+def dict_differences(d1, d2, change):
+	"""returns differences for provided two dictionaries 
+	input d1, d2 type: string/int/float/set/dictionary
+	input change is change type prefix (ex: " -", " +")
+	return value type depends on input d1, d2 type.
+	"""
+	diff = {}
+	if d1 == d2: return None
+	if (not (isinstance(d1, (dict, set)) or isinstance(d2, (dict, set))) and
+		type(d1) != type(d2)): 		
+		raise Exception(f"TypeMismatch- d1:{type(d1)}d2:{type(d2)} - {d1}{d2}")
+	if isinstance(d1, dict):
+		for k_d1, v_d1 in d1.items():
+			if k_d1 not in d2: 
+				diff.update( _get_differences({k_d1: v_d1}, change) )
+				continue
+			if v_d1 == d2[k_d1]: continue
+			diff[k_d1] = dict_differences(v_d1, d2[k_d1], change)
+	elif isinstance(d1, set):
+		diff = _get_differences(d1.difference(d2), change)
+	else:
+		if d1:
+			diff = _get_differences(d1, change)
+
+	return diff
+
+# -----------------------------------------------------------------------------
+#                         Common Dictionary Methods                           #
+# -----------------------------------------------------------------------------
+class DictMethods():
+	"""PAPA DUNDER EXTENSIONS FOR DICTIONARY OBJECTS
+	[self.dic is abstract property which gets iterates over]
+	"""
+
+	def __iter__(self):
+		for k, v in self.dic.items():
+			yield (k, v)
+
+	def __getitem__(self, item):
+		try:
+			return self.dic[item]
+		except KeyError:
+			return None
+
+	def __get__(self, key, item):
+		try:
+			return self[key][item]
+		except KeyError:
+			return None
+
+	def __setitem__(self, item, value):
+		self.dic[item] = value
+
+	def __delitem__(self, srno):
+		try:
+			for k in sorted(self.dic.keys()):
+				if k <= srno: continue
+				self.dic[k-1] = self.dic[k]
+			del(self.dic[k])
+		except:
+			raise KeyError
+
+	def append(self, item, value):
+		"""appends value to self[item] dictionary.
+		create new list if no value found for item, appends to list if available.
+		--> None
+		"""
+		try:
+			if not self.dic.get(item):
+				self.dic[item] = []
+			elif isinstance(self.dic[item], (str, int)):
+				self.dic[item] = [self.dic[item],]
+			self.dic[item].append(value)
+		except:
+			raise Exception
+
 
 # -----------------------------------------------------------------------------
 #                              LOG OPERATIONS                                 #
 # -----------------------------------------------------------------------------
 
 class LOG():
+	"""Collection of static methods for logging.
+	see more...	
+	"""
 
 	@staticmethod
 	def time_stamp():
@@ -823,6 +1084,9 @@ class LOG():
 # -----------------------------------------------------------------------------
 
 class DB():
+	"""Collection of static methods for Database.
+	see more...	
+	"""
 
 	@staticmethod
 	def read_excel(file, sheet='Sheet1', **kwargs):
@@ -842,14 +1106,155 @@ class DB():
 		return pd.read_excel(file, sheet_name=sheet, **kwargs)
 
 
+
+# ------------------------------------------------------------------------------
+# Excel Data WRITE Class, use with context manager
+# ------------------------------------------------------------------------------
+class XL_WRITE():
+	'''EXEL FILE CREATE, 
+	hostname  - excel file name
+	**sht_df    - sht_name=dataframe
+		df        - dataframe which data to be copied to Excel.
+		sht_name  - Sheet Name of Excel where data to be copied
+	Excel O/P file will go inside - ./output/ - path
+	'''
+
+	# Object Initializer
+	def __init__(self, hostname, folder, index=False, **sht_df):
+		i = 0
+		self.hostname = hostname
+		self.folder = folder
+		while True:
+			try:
+				self.__create_excel(hostname, index, **sht_df)
+				break
+			except PermissionError:
+				i += 1
+				hostname = self.hostname+" ("+str(i)+")"
+			except Exception as e:
+				print(e)
+				break
+
+	def __repr__(self): return self.op_file
+
+	# write to Excel/ INTERNAL
+	def __create_excel(self, hostname, index, **sht_df):
+		try:
+			n = 0
+			XL_READ(self.folder + '/'+hostname+'.xlsx', 'tables')
+			while True:
+				n += 1
+				XL_READ(self.folder + '/'+hostname+'-'+str(n)+'.xlsx', 'tables')
+		except:
+			if n == 0:
+				op_file = self.folder + '/'+hostname+'.xlsx'
+			else:
+				op_file = self.folder + '/'+hostname+'-'+str(n)+'.xlsx'
+			self.op_file = op_file
+			with pd.ExcelWriter(op_file) as writer_file:
+				for sht_name, df in sht_df.items():
+					df.to_excel(writer_file, sheet_name=sht_name, index=index)
+
+
+# ------------------------------------------------------------------------------
+# Excel Data Read Class
+# ------------------------------------------------------------------------------
+class XL_READ:
+	'''EXCEL FILE READING,
+	xl      - Excel file to be read
+	shtName - SheetName to be read from given read
+
+	RETURNS
+	-------
+	df      - DataFrame object (iterable, lenth, etc. available )
+
+	USAGE EXAMPLE
+	-------------
+	obj = XL_READ('data/cmd_list.xlsx')					# get xl df Object
+
+	### Length of records ###
+	print(len(obj))
+
+	### Go thru each record ###
+	for header, value in obj:
+		print(header, value)
+
+	### Get a particular column ###
+	print(obj['command'])
+
+	### FILTERING RECORDS ###
+	# Option:1
+	flt = {'level':1, 'dev_type':'cisco_ios'}    # get arguments in a dict
+	x = obj.filter(**flt)                        # apply filter dict
+
+	# Option:2
+	x = obj.filter(dev_type='cisco_ios', level=1)# apply filter manually
+
+	# Option:3 - pass external df for filter
+	x = obj.filter(level=1)                   # created a new DataFrame-x
+	x = obj.filter(df=x, dev_type='cisco_ios')# apply flt manually on new DF-x
+
+	# Option:4 Filter Column & return specific columns only.
+	flt = {'level':1, 'dev_type':'cisco_ios'} # get arguments in a dict
+	col = ['command', 'level']
+	x = obj.column_values(column=col, **flt)
+
+	# Check Output for above options
+	print(x)             # filtered output all columns
+	print(x['xl_col'])   # filtered output with specific col only
+
+	'''
+
+	def __init__(self, xl, shtName='Sheet1'):
+		self.df = pd.read_excel(xl, sheet_name=shtName)
+
+	def __repr__(self):
+		return 'Excel data reprezenting class as DataFrame Object - obj.df'
+
+	def __len__(self): return self.df.last_valid_index()+1
+
+	def __iter__(self):
+		for header, value in self.df.items(): yield (header, value)		
+
+	def __getitem__(self, key): return self.df[key]
+		
+
+	def filter(self, df=None, **kwarg):
+		'''Filter Records
+		df    - external data frame ( default object dataframe )
+		kwarg - filters to be applied on df.
+		'''
+		tmpdf = self.df if df is None else df
+		for k, v in kwarg.items():
+			try:
+				tmpdf = tmpdf[tmpdf[k]==v]
+			except: pass
+		return tmpdf
+
+	def column_values(self, column, **kwarg):
+		'''selected column output, after filters applied
+		column - a single column name or , list of column names
+		kwarg  - filters to be applied
+		'''
+		return self.filter(**kwarg)[column]
+
+
+
+
 # -----------------------------------------------------------------------------
 #                               IP OPERATIONS                                 #
 # -----------------------------------------------------------------------------
 
 class IP():
+	"""Collection of static methods for Networking on (IP).
+	see more...	
+	"""
 
 	@staticmethod
 	def ping_average(ip):
+		"""return average ping responce for provided ip
+		-->int
+		"""
 		lst = popen(f"ping {ip}").read().split("\n")
 		for x in lst:
 			if "Average" in x:
@@ -880,3 +1285,68 @@ class IP():
 			count_of_ones += x.count("1")
 		return 32 - count_of_ones
 
+
+# -----------------------------------------------------------------------------
+#                           Execution secquences                              #
+# -----------------------------------------------------------------------------
+
+class Multi_Execution(Default):
+	"""Template methods for multi-threaded executions.
+	[self.items items are eligible threaded candidates]
+	"""
+
+	max_connections = 30
+
+	def __str__(self): return self._repr()
+
+	def __init__(self, items=None):
+		self.items = items
+
+	def execute_steps(self, multi_thread=True):
+		"""steps defining executions """
+		self.start(multi_thread)
+
+	def start(self, multi_thread=True):
+		"""starting up executins either threaded/sequencial """
+		if not self.items: return None 
+		if multi_thread:
+			self.execute_mt()
+		else: 
+			self.execute_sequencial()
+
+	def end(self):
+		"""Closure process """		
+		pass
+
+	def get_devices(self):
+		"""get devices names from list of files"""
+		self.devices = LST.list_of_devices(self.files)
+
+	def execute_mt(self):
+		"""threaded execution in groups 
+		(self.max_connections defines max threaded processes) 
+		"""
+		for group, items in enumerate(LST.split(self.items, self.max_connections)):
+			self.execute_threads_max(items)
+
+	def execute_threads_max(self, item_list):
+		"""threaded execution of a group
+		"""
+		ts = []
+		for hn in item_list:
+			t = threading.Thread(target=self.execute, args=(hn,) )
+			t.start()
+			ts.append(t)
+		for t in ts: t.join()
+
+	def execute_sequencial(self):
+		"""sequencial execution of items
+		"""
+		for hn in self.items: self.execute(hn)
+
+	@abstractclassmethod
+	def execute(self, hn): 
+		"""abstract class method, to be executed for each item in self.items
+		"""
+		pass
+
